<?php
namespace Drupal\tmgmt_transifex\Transifex;

use Drupal\Component\Serialization\Json;
use Drupal\tmgmt_transifex\Utils\Helpers;
use Drupal\tmgmt\TranslatorInterface;


/**
* Responsible for requests to the Transifex API.
*
* Currently allowed calls have to do with projects, resources, stats and translations.
* The class get all information about the Transifex project and its API settings, the
* moment it gets inittiated.
*/

const ENDPOINT_URL = 'https://rest.api.transifex.com'; # for production

// const ENDPOINT_URL = 'http://apiv3-web:8004'; # for local apiv3

// Low-level doRequest:
//   - Applies options
//   - applies raw request body
//   - returns raw response body
function doRequest($method, $url, $options = array()) {

    $ch = curl_init($url);

    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $method);

    if (isset($options['body'])) {
      curl_setopt($ch, CURLOPT_POSTFIELDS, $options['body']);
    }

    if (isset($options['agent'])) {
      curl_setopt($ch, CURLOPT_USERAGENT, $options['agent']);
    }

    if (isset($options['headers'])) {
      curl_setopt($ch, CURLOPT_HTTPHEADER, $options['headers']);
    }

    if (isset($options['follow'])) {
      curl_setopt($ch, CURLOPT_FOLLOWLOCATION, $options['follow']);
    }

    $response = curl_exec($ch);

    $error = curl_error($ch);
    if ($error) {
      $result = array('error' => $error);
    } else {
      $status = curl_getinfo($ch, CURLINFO_HTTP_CODE);
      $result = array(
        'error' => $status < 200 || $status >= 400,
        'status' => $status,
        'redirect_url' => curl_getinfo($ch, CURLINFO_REDIRECT_URL),
        'body' => $response
      );
    }
    curl_close($ch);
    return $result;
  }



class TransifexApi
{
    const METADATA_NAME = "version";
    const METADATA_NAMESPACE = "__drupal__";
    const POLL_INTERVAL = 5;

    public $translator;

    /**
    * Constructor
    *
    * @param TranslatorInterface $translator
    */
    public function __construct($translator)
    {
        $this->translator = $translator;
    }

    /**
    * Makes a request to the Transifex API.
    *
    * @param string $method The HTTP method
    * @param string $url The URL to make the request
    * @param array $options Array that holds the post data
    */

    // High-level doRequest:
    //   - wraps low-level doRequest
    //   - applies prefix to URL if needed
    //   - applies standard headers
    //   - encodes request body
    //   - returns decoded response body
    function doRequest($method, $url, $options = array()) {
        if (substr($url, 0, strlen(ENDPOINT_URL)) != ENDPOINT_URL) {
        $url = ENDPOINT_URL . $url;
        }
        $options['agent'] = implode(
        " ",
        array(
            "tmgmt-transifex-drupal-version/" . \Drupal::VERSION,
            "tmgmt-transifex-user/" . md5($this->translator->getSetting('api'))
        )
        );
        if (isset($options['body'])) {
            $options['body'] = json_encode($options['body']);
        } else {
            $options['body'] = '';
        }
        if (!isset($options['headers'])) {
            $options['headers'] = array();
        }
        $options['headers'] = array_merge($options['headers'], array(
        'Content-Type: application/vnd.api+json',
        'Content-Length: ' . strlen($options['body']),
        'Authorization: Bearer ' . $this->translator->getSetting('api')
        ));
        $response = doRequest($method, $url, $options);
        if (isset($response['body'])) {
            $response['body'] = json_decode($response['body']);
        }
        return $response;
    }


    /**
    * Get information about a Transifex resource
    *
    * @param $slug
    *
    * @return mixed
    */
    function getResource($slug)
    {
        $response = $this->doRequest('GET', '/resources/' . $this->getResourceId($slug));
        return $response['body']->data->attributes;
    }

    /**
    * Get stats about a Transifex resource
    *
    * @param string $slug The resource's slug
    * @param string $tx_language The langage code to get stats for
    *
    * @return mixed
    */
    function getStats($slug, $tx_language)
    {
        $language_id = $this->getResourceId($slug) . ':l:' . $tx_language;
        $response = $this->doRequest('GET', '/resource_language_stats/' . $language_id);
        return $response['body']->data->attributes;
    }

    /**
    * Get translations for a given Transifex resource
    *
    * @param $slug The resource's slug
    * @param $tx_language$ The langage code to get translations for
    *
    * @return mixed
    */
    function getTranslations($slug, $tx_language)
    {
        $response = $this->doRequest(
            'POST',
            '/resource_translations_async_downloads',
            array('body' => array('data' => array(
              'type' => 'resource_translations_async_downloads',
              'relationships' => array(
                'resource' => array('data' => array(
                  'type' => 'resources',
                  'id' => $this->getResourceId($slug)
                )),
                'language' => array('data' => array(
                  'type' => 'languages',
                  'id' => 'l:' . $tx_language
                ))
              )
            )))
          );
          $url = $response['body']->data->links->self;
          while (true) {
            sleep(self::POLL_INTERVAL);
            $response = $this->doRequest('GET', $url, array('follow' => false));
            if ($response['status'] == 303) {
              // Using low-level doRequest here
              $response = doRequest('GET', $response['redirect_url']);
              return $response['body'];
            }
            if ($response['status'] != 200) {
              throw new TMGMTException("Unable to download file");
            }
            if ($response['body']->data->attributes->status == "failed") {
              throw new TMGMTException("Unable to download file");
            }
        }
    }

    /**
    * Create a new resource in Transifex associated with the given translation
    * job id
    *
    * @param $slug The resource's slug
    * @param $name The resource's name
    * @param $tjid Tmgmt's job id
    * @param $content Resource's content
    *
    * @return mixed|string
    */
    function createResource($slug, $name, $tjid, $content){
        $this->doRequest('POST', '/resources', array('body' => array('data' => array(
            'type' => 'resources',
            'attributes' => array(
              'name' => $name,
              'slug' => $slug,
              'categories' => array('tjid:' . $tjid)
            ),
            'relationships' => array(
              'project' => array('data' => array(
                'type' => 'projects',
                'id' => $this->getProjectId()
              )),
              'i18n_format' => array('data' => array(
                'type' => 'i18n_formats',
                'id' => 'HTML'
              ))
            )
          ))));
          $this->doRequest(
            'POST',
            '/resource_metadata',
            array('body' => array('data' => array(
              'type' => 'resource_metadata',
              'attributes' => array(
                'namespace' => self::METADATA_NAMESPACE,
                'name' => self::METADATA_NAME,
                'value' => \Drupal::VERSION
              ),
              'relationships' => array('resource' => array('data' => array(
                'type' => 'resources',
                'id' => $this->getResourceId($slug)
              )))
            )))
          );
          return $this->_upload_source($slug, $content);
        }

    /**
    * Update an existing resource in Transifex
    *
    * @param $slug The resource's slug
    * @param $name The resource's name
    * @param $categories
    * @param $content Resource's content
    *
    * @return mixed|string
    */
    function updateResource($slug, $name, $categories, $content)
    {
        $this->doRequest(
            'PATCH',
            '/resources/' . $this->getResourceId($slug),
            array('body' => array('data' => array(
              'type' => 'resources',
              'id' => $this->getResourceId($slug),
              'attributes' => array('name' => $name, 'categories' => $categories)
            )))
          );

          $metadata_id = $this->getResourceId($slug) . ':ns:' . self::METADATA_NAMESPACE . ':n:' . self::METADATA_NAME;
          $response = $this->doRequest(
            'PATCH',
            '/resource_metadata/' . $metadata_id,
            array('body' => array('data' => array(
              'type' => 'resource_metadata',
              'id' => $metadata_id,
              'attributes' => array('value' => \Drupal::VERSION)
            )))
          );
          if ($response['status'] == 404) {
            $this->doRequest(
              'POST',
              '/resource_metadata',
              array('body' => array('data' => array(
                'type' => 'resource_metadata',
                'attributes' => array(
                  'namespace' => self::METADATA_NAMESPACE,
                  'name' => self::METADATA_NAME,
                  'value' => \Drupal::VERSION
                ),
                'relationships' => array('resource' => array('data' => array(
                  'type' => 'resources',
                  'id' => $this->getResourceId($slug)
                )))
              )))
            );
          }

          return $this->_upload_source($slug, $content);
        }

        function _upload_source($slug, $content) {
          $response = $this->doRequest(
            'POST',
            '/resource_strings_async_uploads',
            array('body' => array('data' => array(
              'type' => 'resource_strings_async_uploads',
              'attributes' => array(
                'content' => $content,
                'content_encoding' => 'text'
              ),
              'relationships' => array('resource' => array('data' => array(
                'type' => 'resources',
                'id' => $this->getResourceId($slug)
              ))),
            )))

          );
          $url = $response['body']->data->links->self;
          while (true) {
            sleep(self::POLL_INTERVAL);
            $response = $this->doRequest('GET', $url);
            if ($response['status'] != 200) {
              throw new TMGMTException("Unable to upload file");
            }
            $status = $response['body']->data->attributes->status;
            if ($status == 'failed') {
              throw new TMGMTException("Unable to upload file");
            }
            if ($status == 'succeeded') {
              return $response['body']->data->attributes->details;
            }
          }
        }

    /**
    * Create or update resource with the given slug on Transifex
    *
    * @param $slug The resource's slug
    * @param $name The resource's name
    * @param $tjid Tmgmt's job id
    * @param $content Resource's content
    *
    * @return mixed|string
    */
    function upsertResource($slug, $name, $tjid, $content)
    {
        $url = '/resources?filter[project]=' . $this->getProjectId() . '&filter[slug]=' . $slug;
        $ret = $this->doRequest('GET', $url);

        if (count($ret['body']->data) == 0) {
          return $this->createResource($slug, $name, $tjid, $content);
        } else {
          $categories = $ret['body']->data[0]->attributes->categories;
          if ($categories === null || !in_array('tjid:' . $tjid, $categories)) {
            $categories[] = 'tjid:' . $tjid;
          }
          return $this->updateResource($slug, $name, $categories, $content);
        }
    }

    function getProjectId() {
        [$organization_slug, $project_slug] = Helpers::extractSlugFromUrl(
          $this->translator->getSetting('project')
        );
        return 'o:' . $organization_slug . ':p:' . $project_slug;
    }

    /**
    * Return information about a Transifex project
    *
    * @param TranslatorInterface translator
    * @return mixed
    */
    function getProject(TranslatorInterface $translator)
    {
        $tx = new TransifexApi($translator);
        return $tx->doRequest('GET', '/projects/' . $tx->getProjectId());
    }

    function getResourceId($slug) {
        [$organization_slug, $project_slug] = Helpers::extractSlugFromUrl(
          $this->translator->getSetting('project')
        );
        return 'o:' . $organization_slug . ':p:' . $project_slug . ':r:' . $slug;
      }

    function getRemoteTranslations($translator, $resource, $tx_language)
    {
        $content = $this->getTranslations($resource, $tx_language);
        \Drupal::logger('tmgmt_transifex')->info('<pre> '.$content.' </pre>');
        return Helpers::parseRemoteTranslations($content);
    }

}

